import importlib
import inspect
import io
import os
import warnings

import click
import sys

from ward.config import _breakpoint_supported
from ward.terminal import console

original_stdout = sys.stdout


def init_breakpointhooks(pdb_module, sys_module) -> None:
    # breakpoint() is Python 3.7+
    if _breakpoint_supported():
        sys_module.breakpointhook = _breakpointhook
    pdb_module.set_trace = _breakpointhook


def _breakpointhook(*args, **kwargs):
    hookname = os.getenv("PYTHONBREAKPOINT")
    if hookname is None or len(hookname) == 0:
        hookname = "pdb.set_trace"
        kwargs.setdefault("frame", inspect.currentframe().f_back)
    elif hookname == "0":
        return None

    modname, dot, funcname = hookname.rpartition(".")
    if dot == "":
        modname = "builtins"

    try:
        module = importlib.import_module(modname)
        if hookname == "pdb.set_trace":
            set_trace = module.Pdb(stdout=original_stdout, skip=["ward*"]).set_trace
            hook = set_trace
        else:
            hook = getattr(module, funcname)
    except:
        warnings.warn(
            f"Ignoring unimportable $PYTHONBREAKPOINT: {hookname}", RuntimeWarning
        )
        return None

    context = click.get_current_context()
    capture_enabled = context.params.get("capture_output")
    capture_active = isinstance(sys.stdout, io.StringIO)

    if capture_enabled and capture_active:
        sys.stdout = original_stdout
        console.print(f"Entering {modname} - output capturing disabled.", style="info")
        return hook(*args, **kwargs)
    return hook(*args, **kwargs)


__breakpointhook__ = _breakpointhook



import importlib
import importlib.util
import inspect
import os
import pkgutil
from distutils.sysconfig import get_python_lib
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Iterator,
    Collection,
)

import sys
from cucumber_tag_expressions.model import Expression
from importlib._bootstrap import ModuleSpec
from importlib._bootstrap_external import FileFinder

from ward.errors import CollectionError
from ward.fixtures import Fixture
from ward.models import WardMeta
from ward.testing import Test, anonymous_tests, is_test_module_name
from ward.util import get_absolute_path

Glob = str


def is_test_module(module: ModuleType) -> bool:
    return is_test_module_name(module.name)


def _get_module_path(module: ModuleType) -> Path:
    return Path(module.module_finder.find_module(module.name).path)


def _is_excluded_module(module: ModuleType, exclusions: Iterable[Glob]) -> bool:
    return _excluded(_get_module_path(module), exclusions)


def _excluded(path: Path, exclusions: Iterable[Glob]) -> bool:
    """Return True if path matches any of the glob patterns in exclusions. False otherwise."""
    return any(path.match(pattern) for pattern in exclusions)


def _remove_excluded_paths(
    paths: Iterable[Path], exclusions: Iterable[Glob]
) -> List[Path]:
    return [p for p in paths if not _excluded(p, exclusions)]


def _handled_within(module_path: Path, search_paths: Iterable[Path]) -> bool:
    """
    Return True if the module path starts with any of the search paths,
    that is, if any module is contained within any of the search paths
    """
    for search_path in search_paths:
        if search_path.is_dir():
            if search_path in module_path.parents:
                return True
    return False


def get_info_for_modules(
    paths: List[Path], exclude: Tuple[Glob],
) -> Generator[pkgutil.ModuleInfo, None, None]:
    paths = _remove_excluded_paths(set(paths), exclude)

    # Handle case where path points directly to modules
    for path in paths:
        if path.is_file() and not _handled_within(path, paths):
            spec = importlib.util.spec_from_file_location(path.stem, path)
            try:
                mod = importlib.util.module_from_spec(spec)
            except AttributeError as e:
                msg = f"Path {str(path)} is not a valid Python module."
                raise CollectionError(msg) from e
            yield mod

    # Check for modules at the root of the specified path (or paths)
    for mod in pkgutil.iter_modules([str(p) for p in paths if p.is_dir()]):
        if is_test_module(mod) and not _is_excluded_module(mod, exclude):
            yield mod

    # Now check for modules in every subdirectory
    checked_dirs: Set[Path] = set(p for p in paths)
    for p in paths:
        for root, dirs, _ in os.walk(str(p)):
            if _excluded(Path(root), exclude):
                continue
            for dir_name in dirs:
                dir_path = Path(root, dir_name)

                # ignore site-packages directories
                abs_path = dir_path.absolute()
                if str(abs_path).startswith(get_python_lib()):
                    continue

                # if we have seen this path before, skip it
                if dir_path not in checked_dirs and not _excluded(dir_path, exclude):
                    checked_dirs.add(dir_path)
                    for mod in pkgutil.iter_modules([str(dir_path)]):
                        if is_test_module(mod) and not _is_excluded_module(
                            mod, exclude
                        ):
                            yield mod


def load_modules(modules: Iterable[ModuleType]) -> Generator[Any, None, None]:
    for m in modules:
        if hasattr(m, "module_finder"):
            file_finder: FileFinder = m.module_finder
            spec: ModuleSpec = file_finder.find_spec(m.name)
            m = importlib.util.module_from_spec(spec)

        module_name = m.__name__
        if is_test_module_name(module_name):
            if module_name not in sys.modules:
                sys.modules[module_name] = m
            m.__package__ = _build_package_name(m)
            m.__loader__.exec_module(m)
            yield m


def _build_package_name(module: ModuleType) -> str:
    path_without_ext: str = module.__file__.rpartition(".")[0]
    path_with_dots: str = path_without_ext.replace(os.path.sep, ".")
    package_name: str = path_with_dots.rpartition(".")[0]
    return "" if package_name == "." else package_name


def get_tests_in_modules(
    modules: Iterable, capture_output: bool = True
) -> Generator[Test, None, None]:
    for mod in modules:
        mod_name = mod.__name__
        mod_path = get_absolute_path(mod)
        anon_tests: List[Callable] = anonymous_tests[mod_path]
        if anon_tests:
            for test_fn in anon_tests:
                meta: WardMeta = getattr(test_fn, "ward_meta")
                yield Test(
                    fn=test_fn,
                    module_name=mod_name,
                    marker=meta.marker,
                    description=meta.description or "",
                    capture_output=capture_output,
                    tags=meta.tags or [],
                )


def filter_tests(
    tests: Iterable[Test], query: str = "", tag_expr: Optional[Expression] = None,
) -> Iterator[Test]:
    if not query and not tag_expr:
        yield from tests

    for test in tests:
        description = test.description or ""

        matches_query = (
            not query
            or query in description
            or query in f"{test.module_name}."
            or query in inspect.getsource(test.fn)
            or query in test.qualified_name
        )

        matches_tags = not tag_expr or tag_expr.evaluate(test.tags)

        if matches_query and matches_tags:
            yield test


def filter_fixtures(
    fixtures: Iterable[Fixture],
    query: str = "",
    paths: Optional[Collection[Path]] = None,
) -> Iterator[Fixture]:
    if paths is None:
        paths = []
    paths = {path.absolute() for path in paths}

    for fixture in fixtures:
        matches_query = (
            not query
            or query in f"{fixture.module_name}."
            or query in inspect.getsource(fixture.fn)
            or query in fixture.qualified_name
        )

        matches_paths = (
            not paths
            or fixture.path in paths
            or any(parent in paths for parent in fixture.path.parents)
        )

        if matches_query and matches_paths:
            yield fixture



from dataclasses import dataclass
from enum import Enum
from inspect import BoundArguments
from pathlib import Path
from typing import List, Optional, Union, Callable

from ward.errors import FixtureError


class Scope(Enum):
    Test = "test"
    Module = "module"
    Global = "global"

    @classmethod
    def from_str(cls, scope_str: str) -> "Scope":
        try:
            return cls[scope_str.title()]
        except (AttributeError, KeyError) as err:
            raise FixtureError(f"Invalid fixture scope: '{scope_str}'") from err


@dataclass
class Marker:
    name: str
    reason: Optional[str] = None
    when: Union[bool, Callable] = True

    @property
    def active(self):
        try:
            return self.when()
        except TypeError:
            return self.when


@dataclass
class SkipMarker(Marker):
    name: str = "SKIP"


@dataclass
class XfailMarker(Marker):
    name: str = "XFAIL"


@dataclass
class WardMeta:
    marker: Optional[Marker] = None
    description: Optional[str] = None
    tags: Optional[List[str]] = None
    is_fixture: bool = False
    scope: Scope = Scope.Test
    bound_args: Optional[BoundArguments] = None
    path: Optional[Path] = None



import asyncio
import collections
import inspect
from contextlib import suppress
from functools import partial, wraps
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Union,
    Optional,
    Any,
    Generator,
    AsyncGenerator,
    List,
    Iterable,
    Mapping,
    Tuple,
    Collection,
)

from dataclasses import dataclass, field

from ward.models import WardMeta, Scope


@dataclass
class Fixture:
    fn: Callable
    gen: Union[Generator, AsyncGenerator] = None
    resolved_val: Any = None

    def __hash__(self):
        return hash(self._id)

    def __eq__(self, other):
        return self._id == other._id

    @property
    def _id(self):
        return self.__class__, self.name, self.path, self.line_number

    @property
    def key(self) -> str:
        path = self.path
        name = self.name
        return f"{path}::{name}"

    @property
    def scope(self) -> Scope:
        return getattr(self.fn, "ward_meta").scope

    @property
    def name(self):
        return self.fn.__name__

    @property
    def path(self):
        return self.fn.ward_meta.path

    @property
    def module_name(self):
        return self.fn.__module__

    @property
    def qualified_name(self) -> str:
        name = self.name or ""
        return f"{self.module_name}.{name}"

    @property
    def line_number(self) -> int:
        return inspect.getsourcelines(self.fn)[1]

    @property
    def is_generator_fixture(self):
        return inspect.isgeneratorfunction(inspect.unwrap(self.fn))

    @property
    def is_async_generator_fixture(self):
        return inspect.isasyncgenfunction(inspect.unwrap(self.fn))

    @property
    def is_coroutine_fixture(self):
        return inspect.iscoroutinefunction(inspect.unwrap(self.fn))

    def deps(self):
        return inspect.signature(self.fn).parameters

    def parents(self) -> List["Fixture"]:
        """
        Return the parent fixtures of this fixture, as a list of Fixtures.
        """
        return [Fixture(par.default) for par in self.deps().values()]

    def teardown(self):
        # Suppress because we can't know whether there's more code
        # to execute below the yield.
        with suppress(RuntimeError, StopIteration, StopAsyncIteration):
            if self.is_generator_fixture and self.gen:
                next(self.gen)
            elif self.is_async_generator_fixture and self.gen:
                awaitable = self.gen.__anext__()
                asyncio.get_event_loop().run_until_complete(awaitable)


FixtureKey = str
TestId = str
ScopeKey = Union[TestId, Path, Scope]
ScopeCache = Dict[Scope, Dict[ScopeKey, Dict[FixtureKey, Fixture]]]


def _scope_cache_factory():
    return {scope: {} for scope in Scope}


@dataclass
class FixtureCache:
    """
    A collection of caches, each storing data for a different scope.

    When a fixture is resolved, it is stored in the appropriate cache given
    the scope of the fixture.

    A lookup into this cache is a 3 stage process:

    Scope -> ScopeKey -> FixtureKey

    The first 2 lookups (Scope and ScopeKey) let us determine:
        e.g. has a test-scoped fixture been cached for the current test?
        e.g. has a module-scoped fixture been cached for the current test module?

    The final lookup lets us retrieve the actual fixture given a fixture key.
    """

    _scope_cache: ScopeCache = field(default_factory=_scope_cache_factory)

    def _get_subcache(self, scope: Scope) -> Dict[str, Any]:
        return self._scope_cache[scope]

    def get_fixtures_at_scope(
        self, scope: Scope, scope_key: ScopeKey
    ) -> Dict[FixtureKey, Fixture]:
        subcache = self._get_subcache(scope)
        if scope_key not in subcache:
            subcache[scope_key] = {}
        return subcache.get(scope_key)

    def cache_fixture(self, fixture: Fixture, scope_key: ScopeKey):
        """
        Cache a fixture at the appropriate scope for the given test.
        """
        fixtures = self.get_fixtures_at_scope(fixture.scope, scope_key)
        fixtures[fixture.key] = fixture

    def teardown_fixtures_for_scope(self, scope: Scope, scope_key: ScopeKey):
        fixture_dict = self.get_fixtures_at_scope(scope, scope_key)
        fixtures = list(fixture_dict.values())
        for fixture in fixtures:
            with suppress(RuntimeError, StopIteration):
                fixture.teardown()
            del fixture_dict[fixture.key]

    def teardown_global_fixtures(self):
        self.teardown_fixtures_for_scope(Scope.Global, Scope.Global)

    def contains(self, fixture: Fixture, scope: Scope, scope_key: ScopeKey) -> bool:
        fixtures = self.get_fixtures_at_scope(scope, scope_key)
        return fixture.key in fixtures

    def get(
        self, fixture_key: FixtureKey, scope: Scope, scope_key: ScopeKey
    ) -> Fixture:
        fixtures = self.get_fixtures_at_scope(scope, scope_key)
        return fixtures.get(fixture_key)


_DEFINED_FIXTURES = []


def fixture(func=None, *, scope: Optional[Union[Scope, str]] = Scope.Test):
    if not isinstance(scope, Scope):
        scope = Scope.from_str(scope)

    if func is None:
        return partial(fixture, scope=scope)

    # By setting is_fixture = True, the framework will know
    # that if this fixture is provided as a default arg, it
    # is responsible for resolving the value.
    path = Path(inspect.getfile(func)).absolute()
    if hasattr(func, "ward_meta"):
        func.ward_meta.is_fixture = True
        func.ward_meta.path = path
    else:
        func.ward_meta = WardMeta(is_fixture=True, scope=scope, path=path)

    _DEFINED_FIXTURES.append(Fixture(func))

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def using(*using_args, **using_kwargs):
    def decorator_using(func):
        signature = inspect.signature(func)
        bound_args = signature.bind_partial(*using_args, **using_kwargs)
        if hasattr(func, "ward_meta"):
            func.ward_meta.bound_args = bound_args
        else:
            func.ward_meta = WardMeta(bound_args=bound_args)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator_using


def is_fixture(obj: Any) -> bool:
    """
    Returns True if and only if the object is a fixture function
    (it would be False for a Fixture instance,
    but True for the underlying function inside it).
    """
    return hasattr(obj, "ward_meta") and obj.ward_meta.is_fixture


_TYPE_FIXTURE_TO_FIXTURES = Mapping[Fixture, Collection[Fixture]]


def fixture_parents_and_children(
    fixtures: Iterable[Fixture],
) -> Tuple[_TYPE_FIXTURE_TO_FIXTURES, _TYPE_FIXTURE_TO_FIXTURES]:
    """
    Given an iterable of Fixtures, produce two dictionaries:
    the first maps each fixture to its parents (the fixtures it depends on);
    the second maps each fixture to its children (the fixtures that depend on it).
    """
    fixtures_to_parents = {fixture: fixture.parents() for fixture in fixtures}

    # not a defaultdict, because we want to have empty entries if no parents when we return
    fixtures_to_children = {fixture: [] for fixture in fixtures}
    for fixture, parents in fixtures_to_parents.items():
        for parent in parents:
            fixtures_to_children[parent].append(fixture)

    return fixtures_to_parents, fixtures_to_children
